#!perl
use 5.006;
use strict;
use warnings;
use Test::More;

our $VERSION = '0.03';

use GCC::Builtins qw/:all/;

my $res = popcount(2);
my $expected = "1";
is($res, $expected, "called popcount(2) and got result ($res), expected ($expected).");
diag("copy-this-expected-value 'popcount' => '$res',");

done_testing();
